#
#    Uncomplicated VM Builder
#    Copyright (C) 2007-2009 Canonical Ltd.
#    
#    See AUTHORS for list of contributors
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3, as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    The core plugin

import logging
import os
import socket
import tempfile

from   VMBuilder           import register_plugin
import VMBuilder.log       as     log
from   VMBuilder.plugins   import Plugin
from   VMBuilder.exception import VMBuilderUserError

class CorePlugin(Plugin):
    priority = 0

    def register_options(self):
        self.vm.register_setting('-d', '--dest', dest='destdir', help='Specify the destination directory. [default: <hypervisor>-<distro>].')
        self.vm.register_setting('-c', '--config',  type='string', help='Specify a additional configuration file')
        self.vm.register_setting('--debug', action='callback', callback=log.set_verbosity, help='Show debug information')
        self.vm.register_setting('-v', '--verbose', action='callback', callback=log.set_verbosity, help='Show progress information')
        self.vm.register_setting('-q', '--quiet', action='callback', callback=log.set_verbosity, help='Silent operation')
        self.vm.register_setting('-t', '--tmp', default=os.environ.get('TMPDIR', '/tmp'), help='Use TMP as temporary working space for image generation. Defaults to $TMPDIR if it is defined or /tmp otherwise. [default: %default]')
        self.vm.register_setting('--templates', metavar='DIR', help='Prepend DIR to template search path.')
        self.vm.register_setting('-o', '--overwrite', action='store_true', default=False, help='Force overwrite of destination directory if it already exist. [default: %default]')
        self.vm.register_setting('--in-place', action='store_true', default=False, help='Install directly into the filesystem images. This is needed if your $TMPDIR is nodev and/or nosuid, but will result in slightly larger file system images.')
        self.vm.register_setting('--tmpfs', metavar="OPTS", help='Use a tmpfs as the working directory, specifying its size or "-" to use tmpfs default (suid,dev,size=1G).')
        self.vm.register_setting('-m', '--mem', type='int', default=128, help='Assign MEM megabytes of memory to the guest vm. [default: %default]')
        self.vm.register_setting('--cpus', type='int', default=1, help='Number of virtual CPU\'s. [default: %default]')

        group = self.vm.setting_group('Network related options')
        domainname = '.'.join(socket.gethostbyname_ex(socket.gethostname())[0].split('.')[1:]) or "defaultdomain"
        group.add_option('--domain', metavar='DOMAIN', default=domainname, help='Set DOMAIN as the domain name of the guest [default: The domain of the machine running this script: %default].')
        group.add_option('--ip', metavar='ADDRESS', default='dhcp', help='IP address in dotted form [default: %default].')
        group.add_option('--mac', metavar='VALUE', help='MAC address of the guest [default: one will be automatically generated on first run].')
        group.add_option('--mask', metavar='VALUE', help='IP mask in dotted form [default: based on ip setting]. Ignored if --ip is not specified.')
        group.add_option('--net', metavar='ADDRESS', help='IP net address in dotted form [default: based on ip setting]. Ignored if --ip is not specified.')
        group.add_option('--bcast', metavar='VALUE', help='IP broadcast in dotted form [default: based on ip setting]. Ignored if --ip is not specified.')
        group.add_option('--gw', metavar='ADDRESS', help='Gateway (router) address in dotted form [default: based on ip setting (first valid address in the network)]. Ignored if --ip is not specified.')
        group.add_option('--dns', metavar='ADDRESS', help='DNS address in dotted form [default: based on ip setting (first valid address in the network)] Ignored if --ip is not specified.')
        self.vm.register_setting_group(group)

    def preflight_check(self):
        for opt in sum([self.vm.confparser.options(section) for section in self.vm.confparser.sections()], []) + [k for (k,v) in self.vm.confparser.defaults().iteritems()]:
            if '-' in opt:
                raise VMBuilderUserError('You specified a "%s" config option in a config file, but that is not valid. Perhaps you meant "%s"?' % (opt, opt.replace('-', '_')))

    def bootstrap(self):
        """Creates the directory structure where we'll be doing all the work

        When create_directory_structure returns, the following attributes will be set:

         - L{VM.tmproot}: The directory where we build up the guest filesystem

        ..and the corresponding directories are created.

        Additionally, L{VM.destdir} is created, which is where the files (disk images, filesystem
        images, run scripts, etc.) will eventually be placed.
        """

        self.vm.tmproot = self.create_tmpdir()
        logging.debug('Created temporary root: %s', self.vm.tmproot)

        # destdir is where the user's files will land when they're done
        if os.path.exists(self.vm.destdir):
            if self.vm.overwrite:
                logging.info('%s exists, and --overwrite specified. Removing..' % (self.vm.destdir, ))
                shutil.rmtree(self.vm.destdir)
            else:
                raise VMBuilderUserError('%s already exists' % (self.vm.destdir,))

        logging.debug('Creating destination directory: %s', self.vm.destdir)
        os.mkdir(self.vm.destdir)
        self.vm.add_clean_cmd('rmdir', self.vm.destdir)

    def create_tmpdir(self):
        """Creates the working directory for this vm and returns its path"""
        return tempfile.mkdtemp('', 'vmbuilder', self.vm.tmp)

register_plugin(CorePlugin)
