#
#    Uncomplicated VM Builder
#    Copyright (C) 2007-2009 Canonical Ltd.
#    
#    See AUTHORS for list of contributors
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3, as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    The VM class
import ConfigParser
from   gettext             import gettext
import logging
import re
import os
import optparse
import shutil
import tempfile
import textwrap
import socket
import struct
import sys
import urllib
import VMBuilder
import VMBuilder.util              as util
import VMBuilder.log               as log
from   VMBuilder.exception         import VMBuilderException, VMBuilderUserError
_ = gettext

class VM(object):
    """The VM object has the following attributes of relevance to plugins:

    distro: A distro object, representing the distro running in the vm

    disks: The disk images for the vm.
    filesystems: The filesystem images for the vm.

    optparser: Will be of interest mostly to frontends. Any sort of option
               a plugin accepts will be represented in the optparser.

    """
    def __init__(self, conf=None):
        self.distro = None

        self.disks = []
        self.filesystems = []

        self.plugins = []
        self._cleanup_cbs = []

        #: final destination for the disk images
        self.destdir = None
        #: tempdir where we do all the work
        self.workdir = None
        #: directory where we build up the guest filesystem
        self.tmproot = None

        self.optparser = _MyOptParser(epilog="VMBuilder is Copyright (C) 2007-2009 Canonical Ltd. and written by Soren Hansen <soren@canonical.com>.", usage='%prog hypervisor distro [options]')
        self.optparser.arg_help = (('hypervisor', self.hypervisor_help), ('distro', self.distro_help))

        self.confparser = ConfigParser.SafeConfigParser()

        configuration_files = ['/etc/vmbuilder.cfg', os.path.expanduser('~/.vmbuilder.cfg')]

        if conf:
            if not os.path.isfile(conf):
                raise VMBuilderUserError('The path to the configuration file is not valid: %s.' % conf)
            configuration_files += [conf]

        self.confparser.read(configuration_files)

        self.add_clean_cmd('rm', log.logfile)

    def cleanup(self):
        logging.info("Cleaning up")
        while len(self._cleanup_cbs) > 0:
            self._cleanup_cbs.pop(0)()

    def add_clean_cb(self, cb):
        self._cleanup_cbs.insert(0, cb)

    def add_clean_cmd(self, *argv, **kwargs):
        cb = lambda : util.run_cmd(*argv, **kwargs)
        self.add_clean_cb(cb)
        return cb

    def cancel_cleanup(self, cb):
        try:
            self._cleanup_cbs.remove(cb)
        except ValueError, e:
            # Wasn't in there. No worries.
            pass

    def distro_help(self):
        return 'Distro. Valid options: %s' % " ".join(VMBuilder.distros.keys())

    def hypervisor_help(self):
        return 'Hypervisor. Valid options: %s' % " ".join(VMBuilder.hypervisors.keys())

    def register_setting(self, *args, **kwargs):
        return self.optparser.add_option(*args, **kwargs)

    def register_setting_group(self, group):
        return self.optparser.add_option_group(group)

    def setting_group(self, *args, **kwargs):
        return optparse.OptionGroup(self.optparser, *args, **kwargs)

    def add_disk(self, *args, **kwargs):
        """Adds a disk image to the virtual machine"""
        from VMBuilder.disk import Disk
        disk = Disk(*args, **kwargs)
        self.disks.append(disk)
        return disk

    def add_filesystem(self, *args, **kwargs):
        """Adds a filesystem to the virtual machine"""
        from VMBuilder.disk import Filesystem
        fs = Filesystem(*args, **kwargs)
        self.filesystems.append(fs)
        return fs

    def call_hooks(self, func):
        logging.info('Calling hook: %s' % (func,))
        for plugin in self.plugins:
            logging.info('Calling %s method in %s plugin.' % (func, plugin.__module__))
            cbs = list(self._cleanup_cbs)
            getattr(plugin, func)()
        logging.info('Calling %s method in distro plugin.' % (func,))
        getattr(self.distro, func)()
        
    def set_distro(self, arg):
        if arg in VMBuilder.distros.keys():
            self.distro = VMBuilder.distros[arg](self)
            self.set_defaults()
        else:
            raise VMBuilderUserError("Invalid distro. Valid distros: %s" % " ".join(VMBuilder.distros.keys()))

    def get_conf_value(self, key):
        # This is horrible. Did I mention I hate people who (ab)use exceptions
        # to handle non-exceptional events?
        confvalue = None
        try:
            confvalue = self.confparser.get('DEFAULT', key)
        except ConfigParser.NoSectionError, e:
            pass
        except ConfigParser.NoOptionError, e:
            pass

#        try:
#            confvalue = self.confparser.get(self.hypervisor.arg, key)
#        except ConfigParser.NoSectionError, e:
#            pass
#        except ConfigParser.NoOptionError, e:
#            pass

        try:
            confvalue = self.confparser.get(self.distro.arg, key)
        except ConfigParser.NoSectionError, e:
            pass
        except ConfigParser.NoOptionError, e:
            pass

#        try:
#            confvalue = self.confparser.get('%s/%s' % (self.hypervisor.arg, self.distro.arg), key)
#        except ConfigParser.NoSectionError, e:
#            pass
#        except ConfigParser.NoOptionError, e:
#            pass

        logging.debug('Returning value %s for configuration key %s' % (repr(confvalue), key))
        return confvalue
    
    def set_defaults(self):
        """
        is called to give all the plugins and the distro plugin a chance to set
        some reasonable defaults, which the frontend then can inspect and present
        """
        multiline_split = re.compile("\s*,\s*")
        if self.distro:
            for plugin in VMBuilder._plugins:
                self.plugins.append(plugin(self))

            self.optparser.set_defaults(destdir='%s' % (self.distro.arg,))

            (settings, dummy) = self.optparser.parse_args([])
            for (k,v) in settings.__dict__.iteritems():
                confvalue = self.get_conf_value(k)
                if confvalue:
                    if self.optparser.get_option('--%s' % k):
                        if self.optparser.get_option('--%s' % k).action == 'append':
                            values = multiline_split.split(confvalue)
                            setattr(self, k, values)
                        else:
                            setattr(self, k, confvalue)
                    else:
                        setattr(self, k, confvalue)
                else:
                    setattr(self, k, v)

            self.call_hooks('set_defaults')

    def mount_partitions(self):
        """Mounts all the vm's partitions and filesystems below .rootmnt"""
        logging.info('Mounting target filesystems')
        fss = disk.get_ordered_filesystems(self)
        for fs in fss:
            fs.mount()
            self.distro.post_mount(fs)

        self.fsmounted = True

    def umount_partitions(self):
        """Unmounts all the vm's partitions and filesystems"""
        logging.info('Unmounting target filesystem')
        fss = VMBuilder.disk.get_ordered_filesystems(self)
        fss.reverse()
        for fs in fss:
            fs.umount()
        for disk in self.disks:
            disk.unmap()

        self.fsmounted = False

    def installdir(self):
        if self.in_place:
            return self.rootmnt
        else:
            return self.tmproot

    def install(self):
        logging.info("Installing guest operating system. This might take some time...")

        self.call_hooks('post_install')
    
        if not self.in_place:
            logging.info("Copying to disk images")
            util.run_cmd('rsync', '-aHA', '%s/' % self.tmproot, self.rootmnt)

        if self.hypervisor.needs_bootloader:
            logging.info("Installing bootloader")
            self.distro.install_bootloader()

        self.distro.install_vmbuilder_log(log.logfile, self.rootmnt)

    def install_file(self, path, contents=None, source=None, mode=None):
        fullpath = '%s%s' % (self.installdir(), path)
        if source and not contents:
            shutil.copy(source, fullpath) 
        else:
            fp = open(fullpath, 'w')
            fp.write(contents)
            fp.close()
        if mode:
            os.chmod(fullpath, mode)
        return fullpath

    def create(self):
        """
        The core vm creation method
        
        The VM creation happens in the following steps:

        A series of preliminary checks are performed:
         - We check if we're being run as root, since 
           the filesystem handling requires root priv's
         - Each plugin's preflight_check method is called.
           See L{VMBuilder.plugins.Plugin} documentation for details
         - L{create_directory_structure} is called
         - VMBuilder.disk.create_partitions is called
         - VMBuilder.disk.create_filesystems is called
         - .mount_partitions is called
         - .install is called

        """
        util.checkroot()

        failed = False
        try:
            hooks = ['preflight_check',
                     'bootstrap',
                     'customise',
                     'stop',
                     'create_partitions',
                     'create_filesystems',
                     'mount_partitions',
                     'umount_partitions',
                     'finalize',
                     'deploy']

            for hook in hooks:
                if hook == 'stop':
                    sys.exit(0)
                self.call_hooks(hook)

#            util.fix_ownership(self.result_files)

        except VMBuilderException,e:
            failed = True
            raise
        finally:
            if failed:
                logging.debug("Oh, dear, an exception occurred")
            self.cleanup()

class _MyOptParser(optparse.OptionParser):
    def format_arg_help(self, formatter):
        result = []
        for arg in self.arg_help:
            result.append(self.format_arg(formatter, arg))
        return "".join(result)

    def format_arg(self, formatter, arg):
        result = []
        arghelp = arg[1]()
        arg = arg[0]
        width = formatter.help_position - formatter.current_indent - 2
        if len(arg) > width:
            arg = "%*s%s\n" % (self.current_indent, "", arg)
            indent_first = formatter.help_position
        else:                       # start help on same line as opts
            arg = "%*s%-*s  " % (formatter.current_indent, "", width, arg)
            indent_first = 0
        result.append(arg)
        help_lines = textwrap.wrap(arghelp, formatter.help_width)
        result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
        result.extend(["%*s%s\n" % (formatter.help_position, "", line)
                           for line in help_lines[1:]])
        return "".join(result)

    def format_option_help(self, formatter=None):
        if formatter is None:
            formatter = self.formatter
        formatter.store_option_strings(self)
        result = []
        if self.arg_help:
            result.append(formatter.format_heading(_("Arguments")))
            formatter.indent()
            result.append(self.format_arg_help(formatter))
            result.append("\n")
            result.append("*** Use vmbuilder <hypervisor> <distro> --help to get more options. Hypervisor, distro, and plugins specific help is only available when the first two arguments are supplied.\n")
            result.append("\n")
            formatter.dedent()
        result.append(formatter.format_heading(_("Options")))
        formatter.indent()
        if self.option_list:
            result.append(optparse.OptionContainer.format_option_help(self, formatter))
            result.append("\n")
        for group in self.option_groups:
            result.append(group.format_help(formatter))
            result.append("\n")
        formatter.dedent()
        # Drop the last "\n", or the header if no options or option groups:
        return "".join(result[:-1])

