#
#    Uncomplicated VM Builder
#    Copyright (C) 2007-2009 Canonical Ltd.
#    
#    See AUTHORS for list of contributors
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3, as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import unittest

from VMBuilder.plugins.disk import Disk, parse_size, index_to_devname, devname_to_index
from VMBuilder.exception import VMBuilderUserError
from test import VMMock

class TestSizeParser(unittest.TestCase):
    def test_suffixesAreCaseInsensitive(self):
        "Suffixes in size strings are case-insensitive"
        
        for letter in ['K', 'M', 'G']:
            self.assertEqual(parse_size('1%s' % letter), parse_size('1%s' % letter.lower()))
            
    def test_suffixless_counts_as_megabytes(self):
        "Suffix-less size string are counted as megabytes"
        self.assertEqual(parse_size(10), 10)
        self.assertEqual(parse_size('10'), 10)
    
    def test_M_suffix_counts_as_megabytes(self):
        "Sizes with M suffix are counted as megabytes"
        self.assertEqual(parse_size('10M'), 10)
    
    def test_G_suffix_counts_as_gigabytes(self):
        "1G is counted as 1024 megabytes"
        self.assertEqual(parse_size('1G'), 1024)
        
    def test_K_suffix_counts_as_kilobytes(self):
        "1024K is counted as 1 megabyte"
        self.assertEqual(parse_size('1024K'), 1)
        
    def test_rounds_size_to_nearest_megabyte(self):
        "parse_size rounds to nearest MB"
        self.assertEqual(parse_size('1025K'), 1)
        self.assertEqual(parse_size('10250K'), 10)

class TestSequenceFunctions(unittest.TestCase):
    def test_index_to_devname(self):
        self.assertEqual(index_to_devname(0), 'a')
        self.assertEqual(index_to_devname(26), 'aa')
        self.assertEqual(index_to_devname(18277), 'zzz')

    def test_devname_to_index(self):
        self.assertEqual(devname_to_index('a'), 0)
        self.assertEqual(devname_to_index('b'), 1)
        self.assertEqual(devname_to_index('aa'), 26)
        self.assertEqual(devname_to_index('ab'), 27)
        self.assertEqual(devname_to_index('z'), 25)
        self.assertEqual(devname_to_index('zz'), 701)
        self.assertEqual(devname_to_index('zzz'), 18277)

    def test_index_to_devname_and_back(self):
        for i in range(18277):
            self.assertEqual(i, devname_to_index(index_to_devname(i)))

class TestDiskPlugin(unittest.TestCase):
    def setUp(self):
        self.vm = VMMock()

    def test_detect_size(self):
        "detect_size returns the correct size"

        from VMBuilder.util import run_cmd
        from VMBuilder.plugins.disk import detect_size
        run_cmd('qemu-img', 'create', '/tmp/testimg', '5G')
        self.assertTrue(detect_size('/tmp/testimg'), 5*1024)
        imgdev = run_cmd('losetup', '-f', '--show', '/tmp/testimg').strip()
        self.assertTrue(detect_size(imgdev), 5*1024)
        imgdev = run_cmd('losetup', '-d', imgdev)
        os.unlink('/tmp/testimg')

    def test_disk_filename(self):
        self.assertFalse(os.path.exists('foobar.img'))
        disk = Disk(filename='/tmp/foobar.img', size='1G')
        disk.create_image()
        self.assertTrue(os.path.exists('/tmp/foobar.img'))
        os.unlink('/tmp/foobar.img')

    def test_disk_size(self):
        # parse_size only deals with MB resolution
        sizes = [('10G', 10*1024*1024*1024),
                 ('400M', 400*1024*1024),
                 ('345', 345*1024*1024),
                 ('10240k', 10*1024*1024),
                 ('10250k', 10*1024*1024),
                 ('10230k', 10*1024*1024)]

        self.assertFalse(os.path.exists('foobar.img'))
        for (sizestr, size) in sizes:
            disk = Disk(filename='/tmp/foobar.img', size=sizestr)
            disk.create_image()
            actual_size = os.stat('/tmp/foobar.img')[stat.ST_SIZE]
            self.assertEqual(size, actual_size)
        os.unlink('/tmp/foobar.img')

    def test_disk_no_size_given(self):
        self.assertRaises(VMBuilderUserError, Disk, filename='/tmp/randomname')

    def test_existing_image_no_overwrite(self):
        self.assertFalse(os.path.exists('/tmp/foobar.img'))
        fp = open('/tmp/foobar.img', 'w')
        fp.write('canary')
        fp.close()
        disk = Disk(filename='/tmp/foobar.img')
        fp = open('/tmp/foobar.img', 'r')
        self.assertEqual(fp.read(), 'canary')
        fp.close()
        os.unlink('/tmp/foobar.img')

    def test_existing_image_overwrite(self):
        self.assertFalse(os.path.exists('/tmp/foobar.img'))
        fp = open('/tmp/foobar.img', 'w')
        fp.write('canary')
        fp.close()
        disk = Disk(filename='/tmp/foobar.img', size='1G')
        fp = open('/tmp/foobar.img', 'r')
        self.assertNotEqual(fp.read(6), 'canary')
        fp.close()
        os.unlink('/tmp/foobar.img')


